@@ -1,5 +1,10 @@
 from .__dependencies__ import json_fix
 from .__dependencies__ import file_system_py as FS
+from .__dependencies__ import super_map
+from .__dependencies__ import super_hash
+hash_file = super_hash.hash_file
+super_hash = super_hash.super_hash
+LazyDict, Map = super_map.LazyDict, super_map.Map
 
 from time import time as now
 from random import shuffle
@@ -55,42 +60,49 @@ if True:
                 for each_key, each_value in self.__dict__.items():
                     entries += "    "+str(each_key)+" = "+repr(each_value)+",\n"
                 return entries+")"
-
-    def named_list_class(*, names):
+    
+    def create_named_list_class(names):
         """
-            Example:
-                Position = named_list_class(names=['x','y','z'])
-                a = Position([1,2,3])
-                print(a.x)   # 1
-                a.x = 4
-                print(a[0])  # 4
-                a[0] = 9
-                print(a.x)   # 9
+        Example:
+            Position = create_named_list_class(['x','y','z'])
+            a = Position([1,2,3])
+            print(a.x)   # 1
+            a.x = 4
+            print(a[0])  # 4
+            a[0] = 9
+            print(a.x)   # 9
         """
         
+        names_to_index = {}
+        if isinstance(names, dict):
+            names_to_index = names
+        if isinstance(names, (tuple, list)):
+            for index, each in enumerate(names):
+                names_to_index[each] = index
+        
         class NamedList(list):
             def __getitem__(self, key):
-                if isinstance(key, int):
+                if isinstance(key, (int, slice)):
                     return super(NamedList, self).__getitem__(key)
                 # assume its a name
                 else:
                     try:
-                        index = names.index(key)
-                    except ValueError as error:
-                        raise Exception(f'''key={key} not in named list: {self}''')
+                        index = names_to_index[key]
+                    except:
+                        raise KeyError(f'''key={key} not in named list: {self}''')
                     if index >= len(self):
                         return None
                     return self[index]
             
             def __getattr__(self, key):
-                if key in names:
+                if key in names_to_index:
                     return self[key]
                 else:
                     super(NamedList, self).__getattr__(key)
             
             def __setattr__(self, key, value):
-                if key in names:
-                    index = names.index(key)
+                if key in names_to_index:
+                    index = names_to_index[key]
                     while index >= len(self):
                         super(NamedList, self).append(None)
                     super(NamedList, self).__setitem__(index, value)
@@ -102,48 +114,203 @@ if True:
                     super(NamedList, self).__setitem__(key, value)
                 # assume its a name
                 else:
-                    index = names.index(key)
+                    index = names_to_index[key]
                     while index >= len(self):
                         super(NamedList, self).append(None)
                     super(NamedList, self).__setitem__(index, value)
+                    
+            def keys(self):
+                return list(names_to_index.keys())
+            
+            def values(self):
+                return self
+            
+            def get(self, key, default):
+                try:
+                    return self[key]
+                except Exception as error:
+                    return default
+            
+            def items(self):
+                return zip(self.keys(), self.values())
+            
+            def update(self, other):
+                for each_key in names_to_index:
+                    if each_key in other:
+                        self[each_key] = other[each_key]
+                return self
             
             def __repr__(self):
                 import itertools
                 out_string = '['
                 named_values = 0
-                for each_value, each_name in zip(self, names):
-                    named_values += 1
-                    out_string += f' {each_name}={each_value},'
-                # has unnamed values
-                length = len(self)
-                if named_values < length:
-                    for index in itertools.count(named_values-1): # starting where we left off
-                        if index >= length:
-                            break
-                        value = self[index]
-                        out_string += f' {value},'
+                
+                reverse_lookup = {}
+                for each_name, each_index in names_to_index.items():
+                    reverse_lookup[each_index] = reverse_lookup.get(each_index, []) + [ each_name ]
+                    
+                for each_index, value in enumerate(self):
+                    name = "=".join(reverse_lookup.get(each_index, []))
+                    if name:
+                        name += '='
+                    out_string += f' {name}{value},'
                 
                 out_string += ' ]'
                 return out_string
                 
         return NamedList
 
+# 
+# warnings
+# 
+if True:
+    import warnings
+    @singleton
+    class Warnings:
+        _original_filters = list(warnings.filters)
+        _original_showwarning = warnings.showwarning
+        def show_full_stack_trace(self):
+            # show full traceback of each warning
+            import traceback
+            import warnings
+            import sys
+            def warn_with_traceback(message, category, filename, lineno, file=None, line=None):
+                log = file if hasattr(file,'write') else sys.stderr
+                traceback.print_stack(file=log)
+                log.write(warnings.formatwarning(message, category, filename, lineno, line))
+            warnings.showwarning = warn_with_traceback
+            warnings.simplefilter("always")
+        
+        def show_normal(self):
+            warnings.filters = self._original_filters
+            warnings.showwarning = self._original_showwarning
+        
+        def disable(self):
+            warnings.simplefilter("ignore")
+            warnings.filterwarnings('ignore')
+        
+        class disabled:
+            # TODO: in future allow specify which warnings to disable
+            def __init__(with_obj, *args, **kwargs):
+                pass
+            
+            def __enter__(with_obj):
+                with_obj._original_filters = list(warnings.filters)
+                with_obj._original_showwarning = warnings.showwarning
+            
+            def __exit__(with_obj, _, error, traceback):
+                # normal cleanup HERE
+                warnings.filters = with_obj._original_filters
+                warnings.showwarning = with_obj._original_showwarning
+                
+                with_obj._original_filters = list(warnings.filters)
+                with_obj._original_showwarning = warnings.showwarning
+                
+                if error is not None:
+                    raise error
+    # show full stack trace by default instead of just saying "something wrong happened somewhere I guess"
+    Warnings.show_full_stack_trace()
+        
 # 
 # string related
 # 
 if True:
-    def indent(string, by):
-        if isinstance(by, str):
-            indent_string = by
-        else:
-            indent_string = (" "*by)
+    def indent(string, by="    ", ignore_first=False):
+        indent_string = (" "*by) if isinstance(by, int) else by
+        string = string if isinstance(string, str) else stringify(string)
+        start = indent_string if not ignore_first else ""
+        return start + string.replace("\n", "\n"+indent_string)
+
+    def remove_largest_common_prefix(list_of_strings):
+        def all_equal(a_list):
+            if len(a_list) == 0:
+                return True
             
-        return indent_string + f"{string}".replace("\n", "\n"+indent_string)
+            prev = a_list[0]
+            for each in a_list:
+                if prev != each:
+                    return False
+                prev = each
+            
+            return True
+        
+        shortest_path_length = min([ len(each_path) for each_path in list_of_strings ])
+        longest_common_path_length = shortest_path_length
+        while longest_common_path_length > 0:
+            # binary search would be more efficient but its fine
+            longest_common_path_length -= 1
+            if all_equal([ each[0:longest_common_path_length] for each_path in list_of_strings ]):
+                break
+        
+        return [ each[longest_common_path_length:] for each_path in list_of_strings ]
+            
+    def pascal_case_with_spaces(string, valid_word_contents="1234567890qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM"):
+        digits = "1234567890"
+        new_string = " "
+        # get pairwise elements
+        for each_character in string:
+            prev_character = new_string[-1]
+            prev_is_lowercase = prev_character.lower() == prev_character
+            each_is_uppercase = each_character.lower() != each_character
+            
+            # remove misc characters (handles snake case, kebab case, etc)
+            if each_character not in valid_word_contents:
+                new_string += " "
+            # start of word
+            elif prev_character not in valid_word_contents:
+                new_string += each_character.upper()
+            # start of number
+            elif prev_character not in digits and each_character in digits:
+                new_string += each_character
+            # end of number
+            elif prev_character in digits and each_character not in digits:
+                new_string += each_character.upper()
+            # camel case
+            elif prev_is_lowercase and each_is_uppercase:
+                new_string += " "+each_character.upper()
+            else:
+                new_string += each_character
+        
+        # flatten out all the whitespace
+        new_string = new_string.strip()
+        while "  " in new_string:
+            new_string = new_string.replace("  "," ")
+        
+        return new_string
+
+    def levenshtein_distance(s1, s2):
+        # https://stackoverflow.com/questions/2460177/edit-distance-in-python
+        if len(s1) > len(s2):
+            s1, s2 = s2, s1
+        
+        distances = range(len(s1) + 1)
+        for i2, c2 in enumerate(s2):
+            distances_ = [i2+1]
+            for i1, c1 in enumerate(s1):
+                if c1 == c2:
+                    distances_.append(distances[i1])
+                else:
+                    distances_.append(1 + min((distances[i1], distances[i1 + 1], distances_[-1])))
+            distances = distances_
+        return distances[-1]
+    
+    def levenshtein_distance_sort(*, word, other_words):
+        word = word.lower() # otherwise this totally screws up distance
+        prioritized = sorted(other_words, key=lambda each_other: levenshtein_distance(word, each_other))
+        return prioritized
 
 # 
 # generic (any value)
 # 
 if True:
+    def attributes(a_value):
+        if a_value == None:
+            return []
+        all_attachments = dir(a_value)
+        return [
+            each for each in all_attachments if not (each.startswith("__") and each.endswith("__")) and not callable(getattr(a_value, each))
+        ]
+    
     def to_pure(an_object, recursion_help=None):
         # 
         # infinte recursion prevention
@@ -404,6 +571,7 @@ if True:
                 That function will return False until it has been called `size` times
                 Then on the size-th time it returns True, and resets/repeats
         """
+        import time
         if seconds:
             def _countdown():
                     
@@ -545,7 +713,31 @@ if True:
             return_value = tuple(return_value)
         
         return function(return_value, is_key=is_key)
-    
+
+    def all_equal(a_list):
+        if len(a_list) == 0:
+            return True
+        
+        prev = a_list[0]
+        for each in a_list[1:]:
+            if prev != each:
+                return False
+            prev = each
+        
+        return True
+
+    def all_different(a_list):
+        if len(a_list) == 0:
+            return True
+        
+        prev = a_list[0]
+        for each in a_list[1:]:
+            if prev == each:
+                return False
+            prev = each
+        
+        return True
+
 # 
 # math related
 # 
@@ -566,8 +758,36 @@ if True:
         return reduce(operator.mul, iterable, 1)
 
     def max_index(iterable):
+        iterable = tuple(iterable)
+        if len(iterable) == 0:
+            return None
         max_value = max(iterable)
-        return to_pure(iterable).index(max_value)
+        from random import sample
+        options = tuple( each_index for each_index, each in enumerate(iterable) if each == max_value )
+        return sample(options, 1)[0]
+    
+    def max_indices(iterable):
+        iterable = tuple(iterable)
+        if len(iterable) == 0:
+            return None
+        max_value = max(iterable)
+        return tuple( each_index for each_index, each in enumerate(iterable) if each == max_value )
+    
+    def arg_max(*, args, values):
+        values = tuple(values)
+        if len(values) == 0:
+            return None
+        max_value = max(values)
+        from random import sample
+        options = tuple( arg for arg, value in zip(args, values) if value == max_value )
+        return sample(options, 1)[0]
+    
+    def arg_maxs(*, args, values):
+        values = tuple(values)
+        if len(values) == 0:
+            return None
+        max_value = max(values)
+        return tuple( arg for arg, value in zip(args, values) if value == max_value )
 
     def average(iterable):
         from statistics import mean
@@ -651,12 +871,48 @@ if True:
 
 # 
 # time
-# 
-if True:
+#
+if True: 
+    import time
+    
     def unix_time():
-        return int(now()/1000)
-
-
+        return int(time.time()/1000)
+    
+    @singleton
+    class Time:
+        prev = time.time()
+        
+        @property
+        def unix(self):
+            return int(time.time()/1000)
+        
+        @property
+        def time_since_prev_call(self):
+            current = time.time()
+            output = current-prev
+            Time.prev = current
+            return output
+    
+    class Timer:
+        prev = None
+        def __init__(self, name="", *, silence=False, **kwargs):
+            self.name = name
+            self.silence = silence
+        
+        def __enter__(self):
+            self.start_time = int(time.time()/1000)
+            return self
+        
+        def __exit__(self, _, error, traceback):
+            self.end_time = int(time.time()/1000)
+            self.duration = self.end_time - self.start_time
+            if not self.silence:
+                print(f"{self.name} took {self.duration}ms")
+            Timer.prev = self
+            if error is not None:
+                # error cleanup HERE
+                raise error
+    
 # 
 # print helpers
 # 
@@ -926,144 +1182,210 @@ if True:
     # 
     # csv
     # 
-    if True:
-        class Csv:
-            # reads .csv, .tsv, etc
-            @staticmethod
-            def read(path, *, seperator=",", use_headers=None, first_row_is_headers=False, skip_empty_lines=True, comment_symbol=None):
-                import json
-                
-                comments = []
-                rows     = []
-                headers  = []
-                is_first_data_row = True
-                
-                with open(path,'r') as file:
-                    for each_line in file.readlines():
-                        # remove all weird whitespace as a precaution
-                        each_line = each_line.replace("\r", "").replace("\n", "")
-                        
-                        # 
-                        # comments
-                        # 
-                        if isinstance(comment_symbol, str):
-                            if each_line.startswith(comment_symbol):
-                                comments.append(each_line[1:])
-                                continue
-                        
-                        # 
-                        # empty lines
-                        # 
-                        if skip_empty_lines and len(each_line.strip()) == 0:
-                            continue
-                        
-                        # 
-                        # cell data
-                        #
-                        cells = each_line.split(seperator)
-                        cells_with_types = []
-                        skip_to = 0
-                        for index, each_cell in enumerate(cells):
-                            if index < skip_to:
-                                continue
-                            
-                            stripped = each_cell.strip()
-                            if len(stripped) == 0:
-                                cells_with_types.append(None)
-                            else:
-                                first_char = stripped[0]
-                                if not (first_char == '"' or first_char == '[' or first_char == '{'):
-                                    # this converts scientific notation to floats, ints with whitespace to ints, null to None, etc
-                                    try: cells_with_types.append(json.loads(each_cell))
-                                    # if its not valid JSON, just treat it as a string
-                                    except Exception as error:
-                                        cells_with_types.append(each_cell)
-                                else: # if first_char == '"' or first_char == '[' or first_char == '{'
-                                    remaining_end_indicies = reversed(list(range(index, len(cells))))
-                                    skip_to = 0
-                                    for each_remaining_end_index in remaining_end_indicies:
-                                        try:
-                                            cells_with_types.append(
-                                                json.loads(seperator.join(cells[index:each_remaining_end_index]))
-                                            )
-                                            skip_to = each_remaining_index
-                                            break
-                                        except Exception as error:
-                                            pass
-                                    # continue the outer loop
-                                    if skip_to != 0:
-                                        continue
-                                    else:
-                                        # if all fail, go with the default of the shortest cell as a string
-                                        cells_with_types.append(each_cell)
-                        
-                        # 
-                        # headers
-                        # 
-                        if is_first_data_row:
-                            is_first_data_row = False
-                            if first_row_is_headers:
-                                headers.append([ str(each) for each in cells_with_types ])
-                                continue
-                        
-                        rows.append(cells_with_types)
-                
-                # if headers
-                if first_row_is_headers or use_headers:
-                    RowItem = named_list(headers[0] or use_headers)
-                    # tranform each into a named list (backwards compatible with regular list)
-                    rows = [ RowItem(each_row) for each_row in rows ]
-                
-                return comments, headers, rows
+    class Csv:
+        # reads .csv, .tsv, etc 
+        @staticmethod
+        def read(path, *, seperator=",", first_row_is_column_names=False, column_names=None, skip_empty_lines=True, comment_symbol=None):
+            """
+                Examples:
+                    comments, column_names, rows = csv.read("something/file.csv", first_row_is_column_names=True, comment_symbol="#")
+                    comments, _empty_list, rows = csv.read("something/file.csv", first_row_is_column_names=False)
+                    comments, column_names_from_file, rows = csv.read(
+                        "something/file.csv",
+                        column_names=["column1_new_name"],
+                        first_row_is_column_names=True,
+                    )
+                Summary:
+                    Reads in CSV's
+                    - Converts numbers, null, booleans, etc into those types in accordance with JSON
+                    (e.g. null=>None, true=>True, 2.3e31=>float, "hi\n"=>str('hi\n'))
+                    - Anything that is not json-parsable is kept as a string
+                    - Comments can be enabled by with the comment_symbol arg
+                    - Comments must start as the first character of a line, no trailing comments
+                    - Blank spaces (e.g. ,,, ) are converted to None (e.g. ,null,null,)
+                    - Read() will sill parse even if some lines are missing columns
+                Returns:
+                    value: tuple(comments, column_names, rows)
+                    rows:
+                        - Always returns a list
+                        - Each element is a named list
+                        - Named lists inherit from lists (full backwards compatibility)
+                        - Named lists may also be accessed using column_names
+                        for example: rows[0]["column1"] and rows[0].column1 are both valid
+                    column_names:
+                        - Will always return an empty list when first_row_is_column_names=False
+                        - Will always be the column names according to the file (even if overridden)
+                        - Every element in the list will be a string
+                    comments:
+                        - A list of strings
+                        - One string per line
+                        - The comment_symbol itself is removed from the string
+                    
+                Arguments:
+                    path:
+                        - Any string path or path-object
+                        - Will throw error if file does not exist
+                    first_row_is_column_names:
+                        - Boolean, default is False
+                        - If true all elements in the first row will be parsed as strings (even if they look like numbers/null)
+                        - Not all columns need a name
+                        - However using the same name twice or more will cause problems
+                    column_names:
+                        - Optional, a list of strings
+                        - Will override the column_names within the file if provided
+                        - Doesn't need to cover all columns (trailing columns can be unnamed)
+                    skip_empty_lines:
+                        - Boolean, default is True
+                        - A line with spaces or tabs will still qualify as empty
+                    
+            """
+            import json
             
-            @staticmethod
-            def write(*, path, rows, headers=[], comments=[], seperator=",", comment_symbol=None, eol="\n"):
-                import json
-                if len(comments) > 0 and not isinstance(comment_symbol, str):
-                    raise Exception(f'''When writing to a CSV, Csv.write(path={path})\ncomments were given\nbut a comment_symbol was not given\nPlease add a comment_symbol="#" argument or something with a similar string''')
-                
-                def element_to_string(element):
-                    # strings are checked for seperators, if no seperators or whitespace, then unquoted
-                    if isinstance(element, str):
-                        if (
-                            seperator not in element and
-                            eol       not in element and
-                            '\n'      not in element and
-                            '\r'      not in element and (
-                                comment_symbol == None or 
-                                comment_symbol not in element
-                            )
-                        ):
-                            # no need for quoting
-                            return element
-                    # all other values are stored in json format
-                    return json.dumps(element)
-                        
-                
-                with open(path, 'w') as the_file:
+            comments     = []
+            rows         = []
+            file_column_names = []
+            is_first_data_row = True
+            
+            with open(path,'r') as file:
+                for each_line in file.readlines():
+                    # remove all weird whitespace as a precaution
+                    each_line = each_line.replace("\r", "").replace("\n", "")
+                    
                     # 
                     # comments
                     # 
+                    if comment_symbol:
+                        if each_line.startswith(comment_symbol):
+                            comments.append(each_line[len(comment_symbol):])
+                            continue
+                    
+                    # 
+                    # empty lines
+                    # 
+                    if skip_empty_lines and len(each_line.strip()) == 0:
+                        continue
+                    
+                    # 
+                    # cell data
+                    #
+                    cells = each_line.split(seperator)
+                    cells_with_types = []
+                    skip_to = 0
+                    for index, each_cell in enumerate(cells):
+                        if index < skip_to:
+                            continue
+                        
+                        stripped = each_cell.strip()
+                        if len(stripped) == 0:
+                            cells_with_types.append(None)
+                        else:
+                            first_char = stripped[0]
+                            if not (first_char == '"' or first_char == '[' or first_char == '{'):
+                                # this converts scientific notation to floats, ints with whitespace to ints, null to None, etc
+                                try: cells_with_types.append(json.loads(each_cell))
+                                # if its not valid JSON, just treat it as a string
+                                except Exception as error:
+                                    cells_with_types.append(each_cell)
+                            else: # if first_char == '"' or first_char == '[' or first_char == '{'
+                                remaining_end_indicies = reversed(list(range(index, len(cells))))
+                                skip_to = 0
+                                for each_remaining_end_index in remaining_end_indicies:
+                                    try:
+                                        cells_with_types.append(
+                                            json.loads(seperator.join(cells[index:each_remaining_end_index]))
+                                        )
+                                        skip_to = each_remaining_index
+                                        break
+                                    except Exception as error:
+                                        pass
+                                # continue the outer loop
+                                if skip_to != 0:
+                                    continue
+                                else:
+                                    # if all fail, go with the default of the shortest cell as a string
+                                    cells_with_types.append(each_cell)
+                    
+                    # 
+                    # file_column_names
+                    # 
+                    if is_first_data_row:
+                        is_first_data_row = False
+                        if first_row_is_column_names:
+                            file_column_names = [ str(each) for each in cells_with_types ]
+                            continue
+                    
+                    rows.append(cells_with_types)
+            
+            # if file_column_names
+            if first_row_is_column_names or column_names:
+                RowItem = named_list(column_names or file_column_names)
+                # tranform each into a named list (backwards compatible with regular list)
+                rows = [ RowItem(each_row) for each_row in rows ]
+            
+            return comments, file_column_names, rows
+
+        @staticmethod
+        def write(path=None, *, rows=tuple(), column_names=tuple(), seperator=",", eol="\n", comment_symbol=None, comments=tuple()):
+            import json
+            import sys
+            assert comment_symbol or len(comments) == 0, "Comments were provided,"
+            def contains_comment_symbol(string):
+                if not comment_symbol:
+                    return False
+                else:
+                    return comment_symbol in string
+            
+            def element_to_string(element):
+                # strings are checked for seperators, if no seperators or whitespace, then unquoted
+                if isinstance(element, str):
+                    if not (
+                        contains_comment_symbol(element) or
+                        seperator in element or
+                        eol in element or
+                        '\n' in element or
+                        '\r' in element
+                    ):
+                        # no need for quoting
+                        return element
+                # all other values are stored in json format
+                return json.dumps(element)
+            
+            def break_up_comments(comments):
+                for each in comments:
+                    yield from f"{each}".replace("\r", "").split("\n")
+            
+            the_file = sys.stdout if not path else open(path, 'w+')
+            def close_file():
+                if the_file != sys.stdout and the_file != sys.stderr:
+                    try: the_file.close()
+                    except: pass
+            try:
+                # 
+                # comments
+                # 
+                the_file.write(
+                    eol.join([ f"{comment_symbol}{each}" for each in break_up_comments(comments) ])
+                )
+                if len(comments) > 0:
+                    the_file.write(eol)
+                
+                # 
+                # column_names
+                # 
+                if len(column_names) > 0:
                     the_file.write(
-                        eol.join([ f"{comment_symbol}{each}" for each in comments ])
+                        seperator.join(tuple(
+                            element_to_string(str(each)) for each in column_names 
+                        ))+eol
                     )
-                    if len(comments) > 0:
-                        the_file.write(eol)
-                    
-                    # 
-                    # headers
-                    # 
-                    if len(headers) > 0:
-                        the_file.write(
-                            seperator.join(tuple(
-                                element_to_string(str(each)) for each in headers 
-                            ))+eol
-                        )
-                    
-                    # 
-                    # rows
-                    # 
-                    for each_row in rows:
+                
+                # 
+                # rows
+                # 
+                for each_row in rows:
+                    if isinstance(each_row, str):
+                        the_file.write(each_row+eol)
+                    else:
                         row_string_escaped = tuple(
                             element_to_string(each_cell)
                                 for each_cell in each_row 
@@ -1072,6 +1394,12 @@ if True:
                         the_file.write(
                             seperator.join(row_string_escaped)+eol
                         )
+            except Exception as error:
+                # make sure to close the file
+                close_file()
+                raise error
+            
+            close_file()
     
 
 # 
@@ -1093,3 +1421,99 @@ if True:
                 except Exception as error:
                     pass
         return numpy.array(to_pure(value))
+
+# 
+# 
+# Console
+# 
+# 
+@singleton
+class Console:
+    
+    @singleton
+    class foreground:
+        black          = 30
+        red            = 31
+        green          = 32
+        yellow         = 33
+        blue           = 34
+        magenta        = 35
+        cyan           = 36
+        white          = 37
+        bright_black   = 90
+        bright_red     = 91
+        bright_green   = 92
+        bright_yellow  = 93
+        bright_blue    = 94
+        bright_magenta = 95
+        bright_cyan    = 96
+        bright_white   = 97
+    
+    @singleton
+    class background:
+        black          = 40
+        red            = 41
+        green          = 42
+        yellow         = 43
+        blue           = 44
+        magenta        = 45
+        cyan           = 46
+        white          = 47
+        bright_black   = 100
+        bright_red     = 101
+        bright_green   = 102
+        bright_yellow  = 103
+        bright_blue    = 104
+        bright_magenta = 105
+        bright_cyan    = 106
+        bright_white   = 107
+    
+    def color(self, string, foreground="white", background="black"):
+        # if outputing to a file, dont color anything
+        if not sys.stdout.isatty():
+            return string
+        
+        # TODO: detect windows/WSL and disable colors (because powershell and CMD are problematic)
+        
+        foreground_number = getattr(Console.foreground, foreground, None)
+        background_number = getattr(Console.background, background, None)
+        
+        if foreground_number == None:
+            raise Exception(f"couldn't find foreground color {foreground}\nAvailable colors are: {attributes(Console.foreground)}")
+        if background_number == None:
+            raise Exception(f"couldn't find background color {background}\nAvailable colors are: {attributes(Console.background)}")
+        
+        foreground_code = f"\u001b[{foreground_number}m"
+        background_code = f"\u001b[{background_number}m"
+        reset_code = f"\u001b[0m"
+        return foreground_code+background_code+str(string)+reset_code
+
+    class output_redirected_to:
+        def __init__(self, file=None, filepath=None):
+            import sys
+            
+            self.filepath = filepath
+            
+            if self.filepath:
+                FS.ensure_is_folder(self.filepath)
+                self.file = open(self.filepath, "w")
+            else:
+                self.file = file
+                
+            self.real_stdout = sys.stdout
+            self.real_stderr = sys.stderr
+            sys.stdout = self.file
+            sys.stderr = self.file
+        
+        def __enter__(self):
+            return self
+        
+        def __exit__(self, _, error, traceback):
+            if error is not None:
+                import traceback
+                traceback.print_exc()
+            
+            sys.stdout = self.real_stdout
+            sys.stderr = self.real_stderr
+            if self.filepath:
+                self.file.close()
\ No newline at end of file
